from ..generation.rajat_work.qgen.generator.symsub import SymSubGenerator
from ..generation.rajat_work.qgen.generator.fpm.fpm import FPMGenerator
from ..generation.rajat_work.qgen.encoder.dummy import dummyEN
from ..generation.rajat_work.qgen.generator.eda import EDAGenerator
#from ..generation.broken_english.broken_english_generator import BrokenEnglishGen
from ..generation.sentAug.sentAug import AUG
from .generation import GenerateManager

defaultNames  = ["SymSub", "FPM", "EDA", "nlpAug"]
defaultNums = [2,4,4,2]

defaultGenerateManager = GenerateManager(producers= [SymSubGenerator(dummyEN("lite")),FPMGenerator(),EDAGenerator(), AUG()], names= defaultNames,nums= defaultNums)


import spacy

from .base import BaseEncoder


class GloveEncoder(BaseEncoder):
    def __init__(self, model_path):
        super().__init__("Average GloVe", 300, model_path)

        self.nlp = spacy.load(model_path)

    def get_vector(self, sentence):
        return self.nlp(sentence).vector



import codecs
import importlib
import os
import sys


from .base import BaseGenerator

_BATCH_SIZE = 16


class IMTGenerator(BaseGenerator):
    """ Generate questions using text style transfer model trained on iterative matching and translation
        method (Reference: https://arxiv.org/abs/1901.11333) """

    def __init__(self, onmt_path, model_path, n_best):
        """
        :param onmt_path: path to OpenNMT toolkit directory
        :param model_path: path to trained model
        :param n_best: number of rewrites to generate for each sentence
        """
        super().__init__("IMT Style Transfer Model")

        self.n_best = n_best
        self.translator = self._get_translator(onmt_path, model_path, n_best)

    @staticmethod
    def _get_translator(onmt_path, model_path, n_best):
        sys.path.insert(0, onmt_path)
        opts = importlib.import_module('onmt.opts')
        build_translator = importlib.import_module('onmt.translate.translator').build_translator
        argument_parser = importlib.import_module('onmt.utils.parse').ArgumentParser

        onmt_parser = argument_parser()
        opts.translate_opts(onmt_parser)
        opt = onmt_parser.parse_args([f"--model={model_path}", "--src=''", f"--n_best={n_best}", "--replace_unk"])

        return build_translator(opt, report_score=False, out_file=codecs.open(os.devnull, "w", "utf-8"))

    def generate(self, sentence):
        return list(self.batch_generate([sentence]).values())[0]

    def batch_generate(self, sentences):
        _, translation_result = self.translator.translate(sentences, batch_size=_BATCH_SIZE)

        results = dict()
        for i, rewrites in enumerate(translation_result):
            results[sentences[i]] = rewrites

        return results



import os
from typing import List, Tuple, Dict

from numpy.testing._private.utils import assert_raises
from .core.FAQ import Answer, FAQ, Question , FAQUnit, FAQOutput
import numpy as np
from sentence_transformers.readers import InputExample
from sentence_transformers import  SentenceTransformer,SentencesDataset,losses
from .core.exceptions import *
from .modelRelated.utils import cosineSim  , convertForBatchHardTripletLoss
from torch.utils.data import DataLoader
import warnings

class FAQWrapper:
    def __init__(self,id : int ,  FAQ : FAQ):
        self.FAQ = FAQ
        self.id = id
        self.vectors = self.getVectors()


    def getVectors(self) -> np.ndarray:
        faq = self.FAQ.FAQ
        vectors = []
        for unit in faq:
            vector = unit.vectorRep
            if(vector is None):
                raise VectorNotAssignedException()
            vectors.append(vector)
        

        return np.array(vectors)



    def _getClosestQuestions(self,rankedIndices : List[int] ,K : int, topAnswer : str):
        includedSet = set()
        includedSet.add(topAnswer)

        closestQuestions = []
        for ind in rankedIndices:
            currentUnit = self.FAQ.FAQ[ind]
            currentOrignal = currentUnit.orignal.text

            if(currentOrignal not in includedSet):
                includedSet.add(currentOrignal)
                closestQuestions.append(currentOrignal)

            if(len(closestQuestions) == K):
                break

        return closestQuestions

            


        

    def solveForQuery(self,queryVector : np.ndarray, K : int, topSimilar : int = 5)  -> FAQOutput:
        # queryVector has shape (1,emeddingDim)
        if(len(queryVector.shape) == 1):
            queryVector = queryVector.reshape(1,-1)

        assert queryVector.shape[0] == 1 

        cosineScores = cosineSim(queryVector, self.vectors)[0]

        cosineScores = cosineScores.tolist()
        rankedIndices  = [x for x in range(len(cosineScores))]
        rankedIndices.sort(reverse = True, key = lambda x : cosineScores[x])

        maxScore = cosineScores[rankedIndices[0]]
        # Now rankedIndices hold the order of indices with highest to lowest similirity !!!


        competeDict = dict()
        for ind in rankedIndices[:K]:
            # using top K results !!!
            currentlabel = self.FAQ.FAQ[ind].label
            if(currentlabel not in competeDict):
                competeDict[currentlabel] = 0
            competeDict[currentlabel] += cosineScores[ind]

        
        competeList = [(label,score) for label , score in competeDict.items()]
        competeList.sort(key= lambda x : x[1] , reverse= True)

        bestScore = competeList[0][1]
        bestLabel = competeList[0][0]

        bestAnswer = self.FAQ.getAnswerWithLabel(bestLabel)
        bestMatchQuestion = self.FAQ.getQuestionWithLabel(bestLabel)

        return  FAQOutput(faqId= self.id,faqName= self.FAQ.name, answer = bestAnswer,
            question= bestMatchQuestion , score= bestScore,
            similarQuestions=self._getClosestQuestions(rankedIndices,topSimilar,bestMatchQuestion.text) , maxScore= maxScore)
    

        
        




class Bani:
    def __init__(self,FAQs : List[FAQ], modelPath : str = None):
        if(modelPath == None):
            modelPath = 'roberta-base-nli-stsb-mean-tokens'
        self.model : SentenceTransformer = SentenceTransformer(modelPath)
        self.FAQs : List[FAQWrapper] = []
        self.idToFAQ : Dict[int,FAQWrapper] = dict()

        self._registerFAQs(FAQs = FAQs)


    def _registerFAQs(self,FAQs : List[FAQ]):
        """
        registers all the faqs given and then extracts vectors , and forms a 
        gobal index and vector to use for combined question answering 
        """

        assert len(FAQs) > 0
        # All FAQs should be Usable !!!
        for faq in FAQs:
            if(faq.isUsable() == False):
                raise ValueError("All faqs passed to chadBot must be Usable !!!! please build FAQ again or load from preexisting one")
        

        for faq in FAQs:
            if(faq.hasVectorsAssigned()):
                warnings.warn("Vectors already assigned to {} FAQ , if you want to reassign using the current model please clear the vectors using resetAssigned vectors".format(faq.name))
            else:
                print("Assigning vectors to {} faq".format(faq.name))
                faq._assignVectors(model = self.model)


        id = 0
        for faq in FAQs:
            newFAQ = FAQWrapper(id,faq)
            self.FAQs.append(newFAQ)
            self.idToFAQ[id] = newFAQ
            id += 1


    def findClosest(self,query : str,  K : int = 3 , topSimilar : int = 5) -> FAQOutput:
        """
        Here we find the closest from each faq and then compare of the 
        top contenders from different faqs are not dangerouusly similar

        """

        competeList : List[FAQOutput] = []
        queryVector = self.model.encode([query])[0].reshape(1,-1)

        for faq in self.FAQs:
            competeList.append(faq.solveForQuery(queryVector=queryVector, K = K, topSimilar= topSimilar))


        competeList.sort(key = lambda x : x.score, reverse= True)
        # competeList now has answer from each faq in the descending order
        return competeList

         
    def findClosestFromFAQ(self,faqId : int, query : str, K : int = 3, topSimilar : int = 5) -> FAQOutput:
        assert faqId in self.idToFAQ
        faq = self.idToFAQ[faqId]
        queryVector = self.model.encode([query])[0].reshape(1,-1)
        return faq.solveForQuery(queryVector= queryVector, K = K,topSimilar= topSimilar)



    def train(self,outputPath : str, batchSize = 16, epochs : int = 1, **kwargs):
        """
        Trains the model using batch hard triplet loss , 
        for the other kwargs take a look at the documentation for sentencetransformers
        """
        assert batchSize > 4 and epochs > 0 and os.path.exists(outputPath)
        trainingObjectives = [] # training each faq on a different objective
        for faq in self.FAQs:
            trainExamples = convertForBatchHardTripletLoss(faq.FAQ)
            trainDataset = SentencesDataset(trainExamples,self.model)
            trainDataloader = DataLoader(trainDataset, shuffle=True, batch_size= batchSize)
            trainLoss = losses.BatchHardTripletLoss(model= self.model)
            trainingObjectives.append((trainDataloader, trainLoss))

        self.model.fit(train_objectives=  trainingObjectives, warmup_steps= 100,epochs= epochs, save_best_model= False,
            output_path= outputPath, **kwargs)


    def saveModel(self, path):
        os.makedirs(path, exist_ok=True)
        self.model.save(path)


    def saveFAQs(self, rootDirPath : str):
        for faq in self.FAQs:
            coreFaQ = faq.FAQ
            coreFaQ.save(rootDirPath)
    def saveFAQ(self, id : int, rootDirPath : str):
        assert id in self.idToFAQ
        self.idToFAQ[id].FAQ.save(rootDirPath)

    
        







        





       





    









        
        
        


    
    


    




    

    














from sentence_transformers.readers import InputExample
from ..core.FAQ import FAQ
from ..core.exceptions import *
from typing import List
import numpy as np


def cosineSim(v,V):
    """
    computes cosine sim between v,V
    where v and V are  2D np matrices (n,E) (N,E)
    output is of the shape (n,N)
    """

    n1 = np.linalg.norm(v, axis = -1)
    n2 = np.linalg.norm(V, axis = -1)
    dot = np.expand_dims(v,1)*np.expand_dims(V,0)
    # shape (n,N,E)
    dot = dot.sum(axis = -1)
    ans = dot/n1.reshape(-1,1)
    ans = ans/n2.reshape(1,-1)
    return ans



def convertForBatchHardTripletLoss(FAQ : FAQ, minSameLabels = 2) -> List[InputExample]:
    """
    Must check if the data has atleast minSameLabels(atleast 2) examples for each label !!!
    """
    assert minSameLabels >= 2
    faq = FAQ.FAQ

    counter = dict()
    for unit in faq:
        label = unit.label

        if(label not in counter):
            counter[label] = 0
        counter[label] += 1

    for L,count in counter.items():
        if(count < minSameLabels):
            raise TrainDataInvalid("Attempted to use batchhardtriplet loss but label {} has only {} exmaples !!! At least {} required".format(L,count,minSameLabels))
        
    # now converting
    inputs = []
    for unit in faq:
        inputs.append(InputExample(texts = [unit.question.text] , label= unit.label))

    
    return inputs





class WrongInputDataException(Exception):
    pass

class LabelDoesNotExists(Exception):
    pass

class WrongGeneratorOutputException(Exception):
    pass

class VectorNotAssignedException(Exception):
    pass

class WrongModelOutputException(Exception):
    pass

class AttemptedUsingEmptyFAQ(Exception):
    pass

class AttemptedUsingNonUsableFAQ(Exception):
    pass


class TrainDataInvalid(Exception):
    pass


from typing import List, Union, Dict, Tuple
import os
import numpy as np
from .exceptions import *
from .generation import GenerateManager
import pickle







class Sentence:
    def __init__(self, text : str):
        self.text = text

    

class Question(Sentence):
    def __init__(self,text : str, label : int, orignalParaPhrases : List[str] = None):
        super().__init__(text)
        self.label = label
        self.orignalParaPhrases = orignalParaPhrases or []

    def __str__(self) -> str:
        return "{} {}".format(self.text, self.label)


class Answer(Sentence):
    def __init__(self,text : str , label : int):
        super().__init__(text)
        self.label = label


    def __str__(self) -> str:
        return "{} {}".format(self.text, self.label)


class FAQUnit:
    def __init__(self,id : int,question : Question,orignal : Question , answer : Answer):
        """
        ID must be unique for each question(generated or orignal) !!!!
        question is the question itself , if the question is generated from some orignal 
        question then orignal will have the question , otherwise orignal will be same as the question
        answer ofcourse id the answer to the question 
        """
        assert question.label == orignal.label , "label for orignal and current question must be same"
        self.id = id
        self.question = question
        self.orignal = orignal
        self.answer = answer
        self.vectorRep = None  # To be assigned later using a model
        self.label = question.label

    def hasAssignedVector(self):
        if(self.vectorRep is None):
            return False
        return True

    

    def __str__(self) -> str:
        return "Question ---> {}\nAnswer --->{}\n".format(self.question, self.answer)
        
    def __repr__(self) -> str:
        return self.__str__()






def processRaw(questions : List[str], answers : List[str]) -> Tuple[List[Question], List[Answer]]:
    assert len(questions) == len(answers)
    assert len(questions) == len(set(questions)) , "Duplicate questions not allowed"

    
    # Now two or more questions may have the same answer !!! ,How ever it is not recommended !!!
    # You can add your own augmentations later !!!


    a2L = dict()
    label = 0
    for answer in answers:
        if(answer not in a2L):
            a2L[answer] = label
            label += 1
    

    l2Q : Dict[int, Question] = dict()
    for question, answer in zip(questions,answers):
        label = a2L[answer]

        if(label in l2Q):
            # HAndeling if multiple question map to the same answer !!!
            l2Q[label].orignalParaPhrases.append(question)
        else:
            # Creating new question
            l2Q[label] = Question(label= label, text = question)
           # print(label)
           # print(l2Q[label].orignalParaPhrases)



    outAnswers = []
    for answer , label in a2L.items():
        outAnswers.append(Answer(label= label , text=  answer))

    
    outQuestions = list(l2Q.values())

    #for q in outQuestions:
    #    print(len(q.orignalParaPhrases))
    return outQuestions, outAnswers






class FAQ:
    def __init__(self,name : str,questions : List[str] = None, answers : List[str] = None):
        self.name = name
        
        self.questions : List[Question] = None
        self.answers : List[Answer] = None
        

        self.l2Q : Dict[int,Question] = dict()
        self.l2A : Dict[int, Answer] = dict()
        self.FAQ : List[FAQUnit] = None

        if(not (questions is None or answers is None)):
            assert len(questions) == len(answers)
            self.questions, self.answers = processRaw(questions= questions, answers= answers)
            self._runChecks()
            self._buildDicts()



        ######################################################################### 




    def _runChecks(self):
        """
        Running checks on questions and answers

        might add other checks like min/max sentence length etc
        """
        includedSet = set()
        for question in self.questions:
            if(question.label in includedSet):
                raise WrongInputDataException("(ONLY FOR FAQ CLASS ERROR)Label for each question must be unique, found atleast 2 questions with same label")
            includedSet.add(question.label)

        answerlabelsSet = set()
        for answer in self.answers:
            if(answer.label not in includedSet):
                raise WrongInputDataException("Label {} for answer {} does not match any question label , please remove this answer from all answer as it will never be answered anyway !!!".format(answer.label, answer.text))
            
            if(answer.label in answerlabelsSet):
                raise WrongInputDataException("Label for each answer must be unique (the text might be the same !!) but found at least two answers with label {}".format(answer.label))

            answerlabelsSet.add(answer.label)

        
        # checking if any label in question points to no answer at all !!!
        differenceSet = includedSet - answerlabelsSet

        if(len(differenceSet) > 0):
            raise WrongInputDataException("The following questions labels have no corrosponding answer labels {} , please check the input data".format(differenceSet))
        

    def _buildDicts(self):
        """
        To build label to question/answers mappings
        """
        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()



        for question in self.questions:
            self.l2Q[question.label] = question
        
        for answer in self.answers:
            self.l2A[answer.label] = answer




    def getAnswerWithLabel(self, label : int) -> Answer:
        """
        Returns answer for the particular label
        """
        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()

        if(label not in self.l2A):
            raise LabelDoesNotExists("The answer with label {} does not exists".format(label))

        return self.l2A[label]

    
    def getQuestionWithLabel(self, label : int) -> Question:
        """
        Returns all orignal questions for the particular label
        """
        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()
        if(label not in self.l2Q):
            raise LabelDoesNotExists("The question with label {} does not exists".format(label))

        return self.l2Q[label]


    

    
    
    
    def _paraphrase(self, generator : GenerateManager):
        """
        Takes the questions and answers and generated more question using the generator
        given , the number of questions to generate and other settings are to be applied in the 
        generator itself 

        Populated the self.FAQ property 
        """

        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()

        questionTexts= list(map(lambda q : q.text , self.questions))
        generatedQuestions = generator.generate(questionTexts)


        if(len(generatedQuestions) != len(self.questions)):
            raise WrongGeneratorOutputException("number of orignal questions is {} but generator returned {} Lists !!!".format(len(self.questions), len(generatedQuestions)))
        

        
        tempFAQ = []
        idCount = 0
        for orignalQuestion , paraphrases in zip(self.questions, generatedQuestions):
            label = orignalQuestion.label
            answer = self.getAnswerWithLabel(label)
            
            tempFAQ.append(FAQUnit(idCount,orignalQuestion,orignalQuestion,answer= answer))
            idCount += 1

            for parphrase in paraphrases:
                question = Question(label= label , text= parphrase)
                tempFAQ.append(FAQUnit(id= idCount,question= question, orignal= orignalQuestion,answer= answer))
                idCount += 1

            for orignalParaphrase in orignalQuestion.orignalParaPhrases:
                question = Question(label = label , text  =orignalParaphrase)
                tempFAQ.append(FAQUnit(id = idCount, question= question, orignal= orignalQuestion,answer= answer))
                idCount += 1


        self.FAQ = tempFAQ



    def _assignVectors(self,model):
        """
        Using the model to assign vectors to each generated question
        """
        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()
        if(self.isUsable() == False):
            raise AttemptedUsingNonUsableFAQ("Must generate paraphrases before assigning vectors")


        questions = [q.question.text for q in self.FAQ]
        vectors : List[np.ndarray] = model.encode(questions)

        if(len(vectors) != len(questions)):
            raise WrongModelOutputException("The size of list of sentences input to model and that of output do not match , {} != {}".format(len(questions,len(vectors))))

        for i,vector in enumerate(vectors):
            self.FAQ[i].vectorRep = vector
        





    
    def buildFAQ(self,generator : GenerateManager,model = None):
        """
        Will generate questions , and then the vectorrep of each question, 
        WILL NOT SAVE , MUST CALL SAVE 
        """

        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()

        self._paraphrase(generator = generator)
        if(model is not None):
            self._assignVectors(model = model)

      
        



    def isEmpty(self):
        if(self.questions is None or self.answers is None or len(self.questions) == 0 or len(self.answers) == 0):
            return True
        
        return False

    def isUsable(self):
        if(self.isEmpty()):
            return False

        if(self.FAQ is None or len(self.FAQ) == 0):
            return False
        return True

    def hasVectorsAssigned(self):
        if(self.isUsable() == False):
            return False
        for unit in self.FAQ:
            if(unit.hasAssignedVector() == False):
                return False
        return True

        


    def load(self,rootDirPath):
        path = os.path.join(rootDirPath,self.name + ".pkl")
        with open(path, 'rb') as f:
            newObj =  pickle.load(f)
        self.__dict__.update(newObj.__dict__)


    def save(self, rootDirPath):
        path = self,os.path.join(rootDirPath,self.name + ".pkl")
        with open(path, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

        

    def resetAssignedVectors(self, rootDirPath):
        if(self.FAQ is None):
            return
        
        for unit in self.FAQ:
            unit.vectorRep = None
        
        self.save(rootDirPath)


    def resetFAQ(self, rootDirPath):
        self.FAQ = None
        self.l2A = dict()
        self.l2Q = dict()
        self.save(rootDirPath)
        
            

    def __len__(self):
        if(self.isEmpty()):
            raise AttemptedUsingEmptyFAQ()
        if(self.isUsable() == False):
            raise ValueError("cannot access length of a faq whose questions have not been generated !!! MUST CALL BUILD FAQ BEFORE THIS, OR LOAD FAQ FROM PREEXISTING SOURCE")
            
        return len(self.FAQ)





class FAQOutput:
    def __init__(self,answer : Union[Answer,None],question : Question ,faqName : str,faqId : int,score : float, similarQuestions : List[str], maxScore : float):
        """
        Answer == None means the question was out of set
        """
        self.answer = answer
        self.faqId = faqId
        self.faqName = faqName
        self.score = score
        self.similarQuestions = similarQuestions
        self.maxScore = maxScore





    
















    




import importlib
import sys

from .base import BaseGenerator


class ZeroShotGenerator(BaseGenerator):
    """ Generate questions using zero-shot neural machine translation model
        (Reference: https://arxiv.org/abs/1705.07830)"""

    def __init__(self, aqa_path, config_path, model_path):
        """
        :param aqa: path to active-qa directory
        :param model_path: path to trained model
        :param n_best: number of rewrites to generate for each sentence
        """
        super().__init__("ActiveQA Zero-shot neural machine translation model")

        self.reformulator = self._get_translator(aqa_path, config_path, model_path)

    @staticmethod
    def _get_translator(aqa_path, config_path, model_path):
        sys.path.insert(0, aqa_path)
        reformulator = importlib.import_module('px.nmt.reformulator')

        reformulator_instance = reformulator.Reformulator(
            hparams_path=config_path,
            source_prefix='<en> <2en> ',
            out_dir=model_path,
            environment_server_address='')

        return reformulator_instance

    def generate(self, sentence):
        return list(self.batch_generate([sentence]).values())[0]

    def batch_generate(self, sentences):
        reformulator_pb2 = importlib.import_module('px.proto.reformulator_pb2')
        responses = self.reformulator.reformulate(questions=sentences,
                                                  inference_mode=reformulator_pb2.ReformulatorRequest.BEAM_SEARCH)

        results = dict()
        for i, response in enumerate(responses):
            temp = []
            for j, rewrite in enumerate(response):
                temp.append(rewrite.reformulation)
            results[sentences[i]] = temp

        return results



import string
from itertools import product

import numpy as np
from nltk.corpus import stopwords
from nltk.corpus import wordnet as wn
from tqdm import tqdm

from .base import BaseGenerator
import sys

from ..util.nlp import get_spacy_model


class SymSubGenerator(BaseGenerator):
    """ Generate questions via sense-disambiguated synonyms substitution. """

    def __init__(self, encoder, discount_factor=0.5, threshold=0.5):
        """
        :param encoder: encoder for the computation of sentence embeddings
        :param discount_factor: discount factor for weightage calculation during word sense disambiguation (wsd)
        :param threshold: threshold value ranging from 0 to 1 for wsd score.
        """
        super().__init__("Sense-disambiguated Synonym Substitution")

        self.encoder = encoder
        self.discount_factor = discount_factor
        self.threshold = threshold

    def _get_best_sense_key(self, sentence, lemma, pos):
        def _compute_weightage(main_synset, related_synset):
            distance = main_synset.shortest_path_distance(related_synset)
            if distance is None:
                return 1
            else:
                return 1 / (1 + distance)

        wordnet_pos = {'VERB': wn.VERB, 'NOUN': wn.NOUN, 'ADJ': wn.ADJ, 'ADV': wn.ADV}
        if pos not in wordnet_pos:
            return None

        results = []  # tuple of lemma and average similarity score
        lemma_count = 0
        for synset in wn.synsets(lemma, pos=wordnet_pos[pos]):
            if lemma.lower() not in [l.name().lower() for l in synset.lemmas()]:
                continue

            extended_gloss = {synset.definition(): 1}  # gloss and weightage map
            for s in synset.hypernyms():
                extended_gloss[s.definition()] = _compute_weightage(synset, s)
                for e in s.examples():
                    extended_gloss[e] = extended_gloss[s.definition()] * self.discount_factor

            for s in synset.hyponyms():
                extended_gloss[s.definition()] = _compute_weightage(synset, s)
                for e in s.examples():
                    extended_gloss[e] = extended_gloss[s.definition()] * self.discount_factor

            for s in synset.verb_groups():
                extended_gloss[s.definition()] = _compute_weightage(synset, s)
                for e in s.examples():
                    extended_gloss[e] = extended_gloss[s.definition()] * self.discount_factor

            for s in synset.similar_tos():
                extended_gloss[s.definition()] = _compute_weightage(synset, s)
                for e in s.examples():
                    extended_gloss[e] = extended_gloss[s.definition()] * self.discount_factor

            for lemma_ in synset.lemmas():
                if lemma_.name().lower() == lemma.lower():
                    for s in [l.synset() for l in lemma_.derivationally_related_forms()]:
                        extended_gloss[s.definition()] = _compute_weightage(synset, s)
                    break

            for e in synset.examples():
                extended_gloss[e] = 1

            sentences = [sentence] + list(extended_gloss.keys())
            embeddings = self.encoder.get_vectors(sentences)
            similarity_matrix = np.inner(embeddings, embeddings)

            score = 0
            weightage = list(extended_gloss.values())
            for i in range(1, len(similarity_matrix)):
                score += similarity_matrix[0][i] * weightage[i - 1]

            for lemma_ in synset.lemmas():
                if lemma_.name().lower() == lemma.lower():
                    lemma_count += lemma_.count()
                    results.append((lemma_, score))
                    break

        if not results:
            return None

        sense_count = len(results)
        results = [(l.key(), s * ((l.count() + 1) / (lemma_count + sense_count))) for l, s in results]

        return sorted(results, key=lambda r: r[1], reverse=True)[0][0]

    def _get_synonyms(self, spacy_doc, word):
        sentence = spacy_doc.text
        if not sentence or not word or word.lower() in stopwords.words('english') or len(word.split()) > 1:
            return []

        # ignore `word` that is a part of noun_chunk
        for noun_chunk in spacy_doc.noun_chunks:
            tokens = [t.strip(string.punctuation) for t in noun_chunk.text.split()
                      if t.strip(string.punctuation).lower() not in stopwords.words('english')]
            if len(tokens) > 1 and word in tokens:
                return []

        for token in spacy_doc:
            if token.text == word:
                key = self._get_best_sense_key(sentence, token.lemma_, token.pos_)
                if key is None:
                    return []
                else:
                    return [w.replace("_", " ") for w in wn.lemma_from_key(key).synset().lemma_names()
                            if w.lower() != word.lower()]

    def generate(self, sentence):
        return list(self.batch_generate([sentence], use_tqdm=False).values())[0]

    def batch_generate(self, sentences, use_tqdm=True):
        results = dict()
        nlp = get_spacy_model()
        with nlp.disable_pipes('ner'):
            docs = nlp.pipe(sentences)
            for doc in (tqdm(docs, total=len(sentences)) if use_tqdm else docs):
                sentence = doc.text
                tokens = [token.text for token in doc]
                tokens_for_sub = [token for token in tokens if tokens.count(token) == 1]
                token2synonyms = dict()
                for token in tokens_for_sub:
                    syms = self._get_synonyms(doc, token)
                    if syms:
                        token2synonyms[token] = [token] + syms

                result = []
                keys = list(token2synonyms.keys())
                combinations = product(*token2synonyms.values())
                next(combinations)  # skip the first combination as it contains all original tokens
                for combination in combinations:
                    temp = tokens.copy()
                    for i, sub_token in enumerate(combination):
                        key = keys[i]
                        temp[temp.index(key)] = sub_token

                    sent = ' '.join(temp).strip('?').strip() + "?"
                    if sent != sentence:
                        result.append(sent)
                        
                results[sentence] = result

        return results



# Easy data augmentation techniques for text classification
# Jason Wei and Kai Zou
# Original codes adapted from https://github.com/jasonwei20/eda_nlp

import random
import re
from random import shuffle

from nltk.corpus import wordnet
from tqdm import tqdm

from .base import BaseGenerator

random.seed(42)

STOP_WORDS = ['i', 'me', 'my', 'myself', 'we', 'our',
              'ours', 'ourselves', 'you', 'your', 'yours',
              'yourself', 'yourselves', 'he', 'him', 'his',
              'himself', 'she', 'her', 'hers', 'herself',
              'it', 'its', 'itself', 'they', 'them', 'their',
              'theirs', 'themselves', 'what', 'which', 'who',
              'whom', 'this', 'that', 'these', 'those', 'am',
              'is', 'are', 'was', 'were', 'be', 'been', 'being',
              'have', 'has', 'had', 'having', 'do', 'does', 'did',
              'doing', 'a', 'an', 'the', 'and', 'but', 'if', 'or',
              'because', 'as', 'until', 'while', 'of', 'at',
              'by', 'for', 'with', 'about', 'against', 'between',
              'into', 'through', 'during', 'before', 'after',
              'above', 'below', 'to', 'from', 'up', 'down', 'in',
              'out', 'on', 'off', 'over', 'under', 'again',
              'further', 'then', 'once', 'here', 'there', 'when',
              'where', 'why', 'how', 'all', 'any', 'both', 'each',
              'few', 'more', 'most', 'other', 'some', 'such', 'no',
              'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too',
              'very', 's', 't', 'can', 'will', 'just', 'don',
              'should', 'now', '']


class EDAGenerator(BaseGenerator):
    """ Generate questions via Easy Data Augmentation Techniques (Reference: https://arxiv.org/abs/1901.11196). """

    def __init__(self, alpha_sr=0.1, alpha_ri=0.1, alpha_rs=0.1, p_rd=0.1, num_aug=9):
        """
        :param alpha_sr: ratio of words to be replaced by synonyms
        :param alpha_ri: ratio of words to be inserted
        :param alpha_rs: ratio of words to be swapped
        :param p_rd: probability that a word will be deleted
        :param num_aug: number of augmentations
        """
        super().__init__("Easy Data Augmentation Techniques")

        self.alpha_sr = alpha_sr
        self.alpha_ri = alpha_ri
        self.alpha_rs = alpha_rs
        self.p_rd = p_rd
        self.num_aug = num_aug

    @staticmethod
    def _get_only_chars(line):
        clean_line = ""

        line = line.replace("’", "")
        line = line.replace("'", "")
        line = line.replace("-", " ")  # replace hyphens with spaces
        line = line.replace("\t", " ")
        line = line.replace("\n", " ")
        line = line.lower()

        for char in line:
            if char in 'qwertyuiopasdfghjklzxcvbnm ':
                clean_line += char
            else:
                clean_line += ' '

        clean_line = re.sub(' +', ' ', clean_line)  # delete extra spaces
        if clean_line[0] == ' ':
            clean_line = clean_line[1:]
        return clean_line

    @staticmethod
    def _get_synonyms(word):
        synonyms = set()
        for syn in wordnet.synsets(word):
            for l in syn.lemmas():
                synonym = l.name().replace("_", " ").replace("-", " ").lower()
                synonym = "".join([char for char in synonym if char in ' qwertyuiopasdfghjklzxcvbnm'])
                synonyms.add(synonym)
        if word in synonyms:
            synonyms.remove(word)
        return list(synonyms)

    @staticmethod
    def _synonym_replacement(words, n):
        """ Replace n words in the sentence with synonyms from wordnet.
        """
        new_words = words.copy()
        random_word_list = list(set([word for word in words if word not in STOP_WORDS]))
        random.shuffle(random_word_list)
        num_replaced = 0
        for random_word in random_word_list:
            synonyms = EDAGenerator._get_synonyms(random_word)
            if len(synonyms) >= 1:
                synonym = random.choice(list(synonyms))
                new_words = [synonym if word == random_word else word for word in new_words]
                num_replaced += 1
            if num_replaced >= n:  # only replace up to n words
                break

        # this is stupid but we need it, trust me
        sentence = ' '.join(new_words)
        new_words = sentence.split(' ')

        return new_words

    @staticmethod
    def _random_deletion(words, p):
        """ Randomly delete words from the sentence with probability p.
        """
        # obviously, if there's only one word, don't delete it
        if len(words) == 1:
            return words

        # randomly delete words with probability p
        new_words = []
        for word in words:
            r = random.uniform(0, 1)
            if r > p:
                new_words.append(word)

        # if you end up deleting all words, just return a random word
        if len(new_words) == 0:
            rand_int = random.randint(0, len(words) - 1)
            return [words[rand_int]]

        return new_words

    @staticmethod
    def _swap_word(new_words):
        random_idx_1 = random.randint(0, len(new_words) - 1)
        random_idx_2 = random_idx_1
        counter = 0
        while random_idx_2 == random_idx_1:
            random_idx_2 = random.randint(0, len(new_words) - 1)
            counter += 1
            if counter > 3:
                return new_words
        new_words[random_idx_1], new_words[random_idx_2] = new_words[random_idx_2], new_words[random_idx_1]
        return new_words

    @staticmethod
    def _random_swap(words, n):
        """ Randomly swap two words in the sentence n times
        """
        new_words = words.copy()
        for _ in range(n):
            new_words = EDAGenerator._swap_word(new_words)
        return new_words

    @staticmethod
    def _add_word(new_words):
        synonyms = []
        counter = 0
        while len(synonyms) < 1:
            random_word = new_words[random.randint(0, len(new_words) - 1)]
            synonyms = EDAGenerator._get_synonyms(random_word)
            counter += 1
            if counter >= 10:
                return
        random_synonym = synonyms[0]
        random_idx = random.randint(0, len(new_words) - 1)
        new_words.insert(random_idx, random_synonym)

    @staticmethod
    def _random_insertion(words, n):
        """ Randomly insert n words into the sentence
        """
        new_words = words.copy()
        for _ in range(n):
            EDAGenerator._add_word(new_words)
        return new_words

    def generate(self, sentence):
        sentence = self._get_only_chars(sentence)
        words = sentence.split(' ')
        words = [word for word in words if word is not '']
        num_words = len(words)

        augmented_sentences = []
        num_new_per_technique = int(self.num_aug / 4) + 1
        n_sr = max(1, int(self.alpha_sr * num_words))
        n_ri = max(1, int(self.alpha_ri * num_words))
        n_rs = max(1, int(self.alpha_rs * num_words))

        # sr
        for _ in range(num_new_per_technique):
            a_words = self._synonym_replacement(words, n_sr)
            augmented_sentences.append(' '.join(a_words))

        # ri
        for _ in range(num_new_per_technique):
            a_words = self._random_insertion(words, n_ri)
            augmented_sentences.append(' '.join(a_words))

        # rs
        for _ in range(num_new_per_technique):
            a_words = self._random_swap(words, n_rs)
            augmented_sentences.append(' '.join(a_words))

        # rd
        for _ in range(num_new_per_technique):
            a_words = self._random_deletion(words, self.p_rd)
            augmented_sentences.append(' '.join(a_words))

        augmented_sentences = [self._get_only_chars(sentence) for sentence in augmented_sentences]
        shuffle(augmented_sentences)

        # trim so that we have the desired number of augmented sentences
        if self.num_aug >= 1:
            augmented_sentences = augmented_sentences[:self.num_aug]
        else:
            keep_prob = self.num_aug / len(augmented_sentences)
            augmented_sentences = [s for s in augmented_sentences if random.uniform(0, 1) < keep_prob]

        # append the original sentence
        augmented_sentences.append(sentence)

        return augmented_sentences

    def batch_generate(self, sentences):
        results = dict()
        for sentence in tqdm(sentences):
            results[sentence] = self.generate(sentence)

        return results



import nltk

try:
    nltk.data.find("corpora/stopwords")
except LookupError:
    nltk.download('stopwords')
try:
    nltk.data.find("corpora/wordnet")
except LookupError:
    nltk.download('wordnet')



import string

import en_core_web_md
import spacy

_spacy_cache = None
_coref_cache = None

FAST_TOKENIZER = spacy.load('en_core_web_sm', disable=['tagger', 'parser', 'ner', 'textcat'])

# Linguistic Constants
AUXILIARIES = ['am', 'is', 'are', 'was', 'were', 'be', 'could', 'can', 'should', 'shall', 'would', 'will',
               'might', 'may', 'must', 'ought', 'does', 'did', 'do', 'been']
SUBJECTS = ['i', 'me', 'my', 'he', 'him', 'his', 'she', 'her', 'we', 'our', 'us', 'they', 'them', 'their',
            'you', 'your', 'it', 'its']
CONTRACTION = {
    "'tis": "this is",
    "'twas": "this was",
    "ain't": "are not",
    "aren't": "are not",
    "can't": "can not",
    "could've": "could have",
    "couldn't": "could not",
    "didn't": "did not",
    "doesn't": "does not",
    "don't": "do not",
    "hasn't": "has not",
    "he'd": "he had",
    "he'll": "he will",
    "he's": "he is",
    "how'd": "how do",
    "how'll": "how will",
    "how's": "how is",
    "I'd": "I would",
    "I'll": "I will",
    "I'm": "I am",
    "I've": "I have",
    "isn't": "is not",
    "it's": "it is",
    "might've": "might have",
    "mightn't": "might not",
    "must've": "must have",
    "mustn't": "must not",
    "shan't": "shall not",
    "she'd": "she had",
    "she'll": "she will",
    "she's": "she is",
    "should've": "should have",
    "shouldn't": "should not",
    "that'll": "that will",
    "that's": "that is",
    "there's": "there is",
    "they'd": "they had",
    "they'll": "they will",
    "they're": "they are",
    "they've": "they have",
    "wasn't": "was not",
    "we'd": "we had",
    "we'll": "we will",
    "we're": "we are",
    "weren't": "were not",
    "what'd": "what had",
    "what's": "what is",
    "when'd": "when had",
    "when'll": "when will",
    "when's": "when is",
    "where'd": "where had",
    "where'll": "where will",
    "where's": "where is",
    "who'd": "who had",
    "who'll": "who will",
    "who's": "who is",
    "why'd": "why had",
    "why'll": "why will",
    "why's": "why is",
    "won't": "would not",
    "would've": "would have",
    "wouldn't": "would not",
    "you'd": "you had",
    "you'll": "you will",
    "you're": "you are",
    "you've": "you have"
}


def get_spacy_model():
    global _spacy_cache
    if _spacy_cache is None:
        print("Initializing spaCy model...")
        _spacy_cache = en_core_web_md.load()

    return _spacy_cache


def _get_coref_model():
    global _coref_cache

    if _coref_cache is None:
        print("Initializing NeuralCoref model...")
        _coref_cache = spacy.load('en_coref_md', disable=['ner', 'textcat'])

    return _coref_cache


def get_ngrams(sentence, n):
    temp = sentence.split()
    return [' '.join(temp[i:i + n]) for i in range(len(temp) - n + 1)]


def replace_punct(sentence, replace_with="", ignore_list=""):
    """ Replace punctuation in `sentence` with tokens specified by `replace_with`.
    """
    punct = string.punctuation
    for symbol in ignore_list:
        punct = punct.replace(symbol, '')
    if replace_with:
        return sentence.translate(str.maketrans(punct, replace_with * len(punct)))
    else:
        return sentence.translate(str.maketrans('', '', punct))


def convert_contraction(sentence):
    """ Returns converted `sentence` with contradictions replaced. (e.g. "What's" is replaced with "What is")
    """
    tokens = sentence.split()
    for i, t in enumerate(tokens):
        for c in CONTRACTION:
            if t.lower().rstrip(string.punctuation) == c.lower():
                tokens[i] = CONTRACTION[c]
                if t[0].isupper():
                    tokens[i] = tokens[i].capitalize()

    return ' '.join(tokens)


def resolve_coref(sentence, context):
    """ Return a modified sentence with "mentions" replaced by the main entity
    """

    # this is only a temporaty fix !!!!!

    return sentence


    coref_model = _get_coref_model()
    doc = coref_model(context)
    if not doc._.has_coref:
        return sentence
    else:
        # get the phrase surrounding the mention
        sent_tokens = [token.text for token in FAST_TOKENIZER(sentence)]
        for cluster in doc._.coref_clusters:
            for span in cluster.mentions:
                if cluster.main.text.lower() in SUBJECTS or span.text.lower() == cluster.main.text.lower() \
                        or cluster.main.text in sentence:
                    # Don't resolve coreference if any of the following is true:
                    #    1. Cluster.main.text (the phrase which the mention is resolved to) is in the SUBJECTS list,
                    #       since they are not what we're looking for (e.g. mention = "a boy", Cluster.main.text = "he")
                    #    2. Mention and Cluster.main.text are the same
                    #    3. Cluster.main.text is already in the output sentence
                    continue

                if span.text.lower() == 'it':  # heuristic to gather more context if mention is the word 'it'
                    index = span.start
                    if index == len(
                            doc) - 1:  # if mention is the last word in doc, we take the last 3 tokens as phrase
                        phrase = doc[index - 2:].text
                    else:  # else we take one word before the mention and one word after the mention to form a phrase
                        phrase = doc[index - 1:index + 2].text
                else:
                    phrase = span.text

                phrase = phrase.strip(string.punctuation)  # remove leading and ending punctuation
                phrase_len = len(phrase.split())

                # check if phrase is in sent_doc
                for j in range(len(sent_tokens)):
                    if " ".join(sent_tokens[j: j + phrase_len]) == phrase:
                        # delete mention and insert the main entity
                        start = j + sent_tokens[j:].index(span[0].text)
                        end = j + sent_tokens[j:].index(span[-1].text)
                        sent_tokens = sent_tokens[:start] + sent_tokens[end + 1:]
                        sent_tokens.insert(start, cluster.main.text)
                        break

        return " ".join(sent_tokens)



from .base import BaseEncoder
import numpy as np

class dummyEN(BaseEncoder):
    def __init__(self, model_path):
        super().__init__("dummy", 300, "lite")

    def get_vector(self, sentence):
        return np.random.rand(300)

    def get_vectors(self,sentences):
        return np.random.rand(len(sentences),100)


import re
from collections import defaultdict
from itertools import product

import regex

from  ...util import nlp
from .pattern import TOKENS as SPECIAL_TOKENS

_spacy_docs_cache = dict()


class MatchedResult:
    def __init__(self, group_id, pattern, match):
        self.group_id = group_id
        self.pattern = pattern
        self.fuzzy_counts = sum(match.fuzzy_counts)
        self.tokens = self._parse_tokens(match.string, pattern, match)

    @staticmethod
    def _get_spacy_doc(sentence):
        spacy_nlp = nlp.get_spacy_model()
        if sentence not in _spacy_docs_cache:
            with spacy_nlp.disable_pipes('ner'):
                _spacy_docs_cache[sentence] = spacy_nlp(sentence)

        return _spacy_docs_cache[sentence]

    @staticmethod
    def _get_first_verb_offset(sentence, starting_offset=0):
        """ Return character offset for the first encountered verb
        """
        doc = MatchedResult._get_spacy_doc(sentence)
        for token in doc:
            if token.idx < starting_offset:
                continue
            # single verb or adverb-verb pair
            elif token.pos_ == 'VERB' or (
                    token.tag_ == 'RB' and (token.i + 1 < len(doc) and doc[token.i + 1].pos_ == 'VERB')
            ):
                return token.idx

        return -1

    @staticmethod
    def _is_subj(sentence, substr):
        """ Returns True if `substr` in `sentence` is a noun chunk and its syntactic dependency tag is nominal subject
        """
        # remove punctuations to make things easier
        sentence = " ".join(nlp.replace_punct(sentence, replace_with=" ", ignore_list="-").strip().split())
        substr = " ".join(nlp.replace_punct(substr, replace_with=" ", ignore_list="-").strip().split())

        doc = MatchedResult._get_spacy_doc(sentence)
        temp = []
        sbj_tags = ['nsubj', 'nsubjpass']
        previous_dep = None
        for chunk in doc.noun_chunks:
            if chunk.text == substr and chunk.root.dep_ in sbj_tags:
                return True
            elif chunk.text in nlp.get_ngrams(substr, len(chunk.text.split())):
                # sometimes the subject might be combination of chunks, so we're keeping this chunk for later
                # *Note: Consecutive subject chunk is discarded since they are two separate and not
                # related noun chunks. (i.e. If they belongs to same chunk, they'd be grouped together as
                # a chunk, not as 2 separate chunks)
                if not (previous_dep in sbj_tags and chunk.root.dep_ in sbj_tags):
                    temp.append(chunk.text)
                    previous_dep = chunk.root.dep_

        if temp and substr.startswith(temp[0]) and substr.endswith(temp[-1]):
            return True

        return False

    @staticmethod
    def _parse_tokens(sentence, pattern, match):
        tokens = defaultdict(list)
        for t in SPECIAL_TOKENS:
            if t in pattern:
                tokens[t] = match.captures(t[1:-1])

        # resolve <sbj> <act> and <obj> <act>
        if "<sbj> <act>" in pattern or "<obj> <act>" in pattern:
            pairs = list(product(tokens['<sbj>'] if "<sbj> <act>" in pattern else tokens['<obj>'], tokens['<act>']))
            unresolved = []
            for pair in pairs:
                if " ".join(pair) in sentence:
                    unresolved.append(pair)

            for pair in unresolved:
                sub_sentence = " ".join(pair)
                verb_offset = MatchedResult._get_first_verb_offset(sentence, sentence.index(sub_sentence))
                if verb_offset == -1:
                    tokens = defaultdict(list)
                    break

                second_part = sentence[verb_offset:sentence.index(sub_sentence) + len(sub_sentence)]
                first_part = sub_sentence[:sub_sentence.rfind(second_part)]

                if pair[0] in tokens['<sbj>']:
                    index = tokens['<sbj>'].index(pair[0])
                    tokens['<sbj>'][index] = first_part
                else:
                    index = tokens['<obj>'].index(pair[0])
                    tokens['<obj>'][index] = first_part

                act_index = tokens['<act>'].index(pair[1])
                tokens['<act>'][act_index] = second_part

        # Verify <sbj> tokens
        for sbj in tokens['<sbj>']:
            if not MatchedResult._is_subj(sentence, sbj):
                tokens = defaultdict(list)

        return tokens


class FuzzyMatcher:
    def __init__(self, patterns):
        self.patterns = [[(pattern, self._convert_to_fuzzy_regex(pattern)) for pattern in group] for group in patterns]

    @staticmethod
    def _convert_to_fuzzy_regex(pattern):
        match_any = '.+'
        fuzzy_rule = '{e<=3}'
        ignore_case = '(?i)'
        space = re.escape(' ')
        placeholder = '<<<placeholder>>>'

        special_tokens = [token for token in pattern.split() if token in SPECIAL_TOKENS]
        pattern = re.escape(pattern)
        for token in SPECIAL_TOKENS:
            pattern = pattern.replace(re.escape(token), match_any)

        parts = pattern.split(match_any)
        for i in range(len(parts)):
            part = parts[i]
            if len(part.strip(space)) != 0:
                add_start = part.startswith(space)
                add_end = part.endswith(space)
                parts[i] = f"{space if add_start else ''}({part.strip(space)}){fuzzy_rule}{space if add_end else ''}"

        fuzzy_regex = rf"{ignore_case}{match_any.join(parts)}"
        # Add named capture groups
        for sp in special_tokens:
            fuzzy_regex = fuzzy_regex.replace(match_any, f"(?{sp}{placeholder})", 1)
        fuzzy_regex = fuzzy_regex.replace(placeholder, match_any)

        return fuzzy_regex

    def match(self, sentence):
        def get_effective_length(_pattern):
            # get pattern length without counting special tokens
            _pattern = _pattern.replace("<sbj>", "")
            _pattern = _pattern.replace("<obj>", "")
            _pattern = _pattern.replace("<act>", "")
            _pattern = _pattern.replace("<st>", "")

            return len(_pattern)

        candidates = []
        for group_id, group in enumerate(self.patterns):
            for pattern, fuzzy_regex in group:
                match = regex.fullmatch(fuzzy_regex, sentence.strip("?"))
                if match is not None:
                    candidates.append(MatchedResult(group_id, pattern, match))

        # sort by least fuzzy count, then by most pattern length
        candidates.sort(key=lambda x: (x.fuzzy_counts, -get_effective_length(x.pattern)))

        return candidates[0] if len(candidates) > 0 else None



from common.utils import *
from common.FAQ import *
from sentence_transformers import  SentenceTransformer

from common.generation import IdentityProducer


from common.generation import GeneratonManager
q2L = load_dict("babybonus/question_to_label.pkl")
a2L = load_dict("babybonus/answer_to_label.pkl")



l2A = dict()
for a,l in a2L.items():
    l2A[l] = a


questions = []
answers = []

limit = 1000
for q,l in q2L.items():
    questions.append(q)
    answers.append(l2A[l])
    if(limit < 0):
        break
    limit -= 1


Q,A = processRaw(questions,answers)


faq = FAQ(name = "babyBonus", questions= Q, answers= A)
model  = SentenceTransformer("roberta-base-nli-stsb-mean-tokens")
#faq._paraphrase(Identitygenerator())

faq.buildFAQ(model ,GeneratonManager([IdentityProducer]))
print(len(faq.FAQ))
#faq.save('./')
print(faq.isUsable())


"""

#generationManager = GeneratonManager(producers= [SymSubGenerator(dummyEN("lite")),FPMGenerator()])

#print(generationManager.generate(questions[:10]))

"""


class BaseGenerator:
    def __init__(self, name):
        self.name = name

    def generate(self, sentence):
        raise NotImplementedError("'generate' is not implemented")

    def batch_generate(self, sentences):
        raise NotImplementedError("'batch_generate' is not implemented")



import fasttext

from .base import BaseEncoder


class FTEncoder(BaseEncoder):
    def __init__(self, model_path):
        super().__init__("fastText", 300, model_path)

        self.model = fasttext.load_model(self.model_path)

    def get_vector(self, sentence):
        return self.model.get_sentence_vector(sentence)



# **************************************************************************************************
# Tokens and their meaning:
# 1. <obj>  : object (includes time and place)
# 2. <sbj>  : subject
# 3. <act>  : action (must begin with verb)
# 4. <st>   : statement/short phrase describing a situation (can form complete sentence by itself)
# **************************************************************************************************

TOKENS = ['<sbj>', '<obj>', '<act>', '<st>']

pattern_specs = [
    # yes/no question (existential)
    {
        "patterns": [
            "{aux1} there <obj>",
            "{aux1} there any <obj>",
            "{aux2} there be <obj>",
            "{aux2} there be any <obj>",
            "Any <obj>"
        ],
        "substitution_keys": {"aux1": ['is', 'are'], "aux2": ['will', 'would']},

        "description": "yes/no question (existential)"
    },

    # yes/no question (obligatory)
    {
        "patterns": [
            "Is it {necessary} for <sbj> to <act>",
            "Is it {necessary} to <act>",
            "{pre-if} <st> Is it {necessary} for <sbj> to <act>",
            "Is it {necessary} for <sbj> to <act> {if} <st>",
            "{aux1} <sbj> {required} to <act>",
            "{pre-if} <st> {aux1} <sbj> {required} to <act>",
            "{aux2} <sbj> {need} to <act>",
            "{aux2} <sbj> {need} to <act> {if} <st>",
            "{pre-if} <st> {aux2} <sbj> {need} to <act>",
            "{aux3} <sbj> <act>",
            "{pre-if} <st> {aux3} <sbj> <act>",
            "<st> {aux3} <sbj> <act>",
        ],
        "substitution_keys": {
            "necessary": ['necessary', 'compulsory', 'required', 'obligatory'],
            "required": ['required', 'obliged'],
            "if": ['if', 'when'],
            "pre-if": ['if', 'when', ''],
            "need": ["have", "need", "still need", "still have"],
            "aux1": ['is', 'am', 'are'],
            "aux2": ['do', 'does'],
            "aux3": ['must', 'should', 'shall']
        },

        "description": "yes/no question (obligatory)"
    },

    # yes/no question (location)
    {
        "patterns": [
            "{aux} <st> possible {within} <obj>",
        ],
        "substitution_keys": {"aux": ['is', 'are'], "within": ['within', 'inside', 'in']},

        "description": "yes/no question (location)"
    },

    # generic yes/no question (conditional)
    {
        "patterns": [
            "{aux} <st>",
            "Am I right to say that <st>",

            "<st> {true}",
        ],
        "substitution_keys": {
            "aux": ['do', 'does'],
            "true": ["True", "True or False", "Really", "Ya meh", "Is it true", "True or not"]
        },

        "description": "generic yes/no question (conditional)"
    },

    # yes/no question (conditional)
    {
        "patterns": [
            "If <st> {aux1} <sbj> <act>",
            "<st> {aux1} <sbj> <act>",
            "{aux2} there <obj> where <st>",
            "{aux1} <sbj> <act> if <st>",
            "{aux3} there be <obj> if <st>",

            "For <st> {aux2} there <obj>",
            "For <st> got <obj>",
            "Given <st> got <obj>",
        ],
        "substitution_keys": {"aux1": ['do', 'does'], "aux2": ['is', 'are'], "aux3": ['will', 'would']},

        "description": "yes/no question (conditional)"
    },

    # yes/no question (post-conditional)
    {
        "patterns": [
            "After <st> {aux1} <sbj> {need} to <act>",

            "{aux1} <sbj> {need} to <act> after <st>",
        ],
        "substitution_keys": {"aux1": ['do', 'does'], "need": ['have', 'still have', 'need', 'still need']},

        "description": "yes/no question (post-conditional)"
    },

    # can/cannot question
    {
        "patterns": [
            "{aux1} <sbj> <act>",
            "{aux2} <sbj> {allowed} to <act>",
            "Is it {possible} for <sbj> to <act>",
            "Is it {possible} if <sbj> <act>",
            "Is it {possible} to <act>",
            "{aux2} <sbj> able to <act>",
            "Can <st>",
        ],
        "substitution_keys": {
            "allowed": ['allowed', 'entitled'],
            "possible": ['possible', 'alright', 'ok'],
            "aux1": ['can', 'could', 'may'],
            "aux2": ['am', 'is', 'are'],
        },

        "description": "can/cannot question"
    },

    # can/cannot question (two actions)
    {
        "patterns": [
            "{aux1} <sbj> <act> {to} <act>",
        ],
        "substitution_keys": {
            "aux1": ['can', 'could'],
            "to": ['to be', 'which is to be'],
        },

        "description": "can/cannot question (two actions)"
    },

    # can/cannot question (role play)
    {
        "patterns": [
            "{aux1} <sbj> working as <obj> <act>",

            "As <obj> , {aux1} <sbj> <act>",
        ],
        "substitution_keys": {"aux1": ['can', 'could']},

        "description": "can/cannot question (role play)"
    },

    # can/cannot question (conditional)
    {
        "patterns": [
            "<st> Is it possible for <sbj> to <act>",
            "{if} <st> Is it possible for <sbj> to <act>",
            "<st> {aux1} <sbj> <act>",
            "{if} <st> {aux1} <sbj> <act>",
            "{aux1} <sbj> <act> {if} <st>",
            "<st> May I know if <obj> {aux1} <act>",
            "{if} <st> May I know if <obj> {aux1} <act>",
            "{aux2} <sbj> allowed to <act> if <st>",
            "I understand that <st> but {aux1} <sbj> <act>",
            "I understand that <st> However, {aux1} <sbj> <act>",
        ],
        "substitution_keys": {
            "aux1": ['can', 'could', 'may'],
            "aux2": ['am', 'is', 'are'],
            "if": ['if', 'when', 'while'],
        },

        "description": "can/cannot question (conditional)"
    },

    # can/cannot question (post-conditional)
    {
        "patterns": [
            "Is it possible to <act> {after} <st>",
            "{aux} <sbj> <act> {after} <st>",
        ],
        "substitution_keys": {
            "aux": ['can', 'could'],
            "after": ['after', 'now that'],
        },

        "description": "can/cannot question (post-conditional)"
    },

    # can it be question
    {
        "patterns": [
            "{aux} <sbj> be <act>",
            "{aux} <sbj> be <obj>",

            "{aux} <obj> be <sbj>",
        ],
        "substitution_keys": {"aux": ['can', 'could']},

        "description": "can it be question"
    },

    # what question
    {
        "patterns": [
            "What {aux1} <obj>",
            "<st> What {aux1} <obj>",
            "{description} <obj>",
            "What {aux2} <obj> <act>",
            "What {aux1} <obj> all about",
            "What is meant by <obj>",
            "What {aux2} <obj> mean",
            "What <obj> {aux2} <sbj> {need} to <act>",
            "<st> What <obj> {aux2} <sbj> {need} to <act>",
            "What <obj>"
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "description": ["Can you tell me more about",
                            "Can you tell me the details about",
                            "May I know about",
                            "I want to know about",
                            "Explain to me"],
            "need": ['need', 'required']
        },

        "description": "what question"
    },

    # what question (capability)
    {
        "patterns": [
            "What {aux1} <sbj> <act>",
            "What <obj> {aux1} <sbj> <act>",
            "What <obj> {aux1} <sbj> <act> as <st>",
        ],
        "substitution_keys": {
            "aux1": ['can', 'could', 'should', 'shall', 'do', 'does'],
            "aux2": ['do', 'does'],
        },

        "description": "what question (capability)"
    },

    # what question (location)
    {
        "patterns": [
            "What {aux} <sbj> do at <obj>",
        ],
        "substitution_keys": {"aux": ['can', 'could']},

        "description": "what question (location)"
    },

    # what question (example)
    {
        "patterns": [
            "{description} <obj>",
        ],
        "substitution_keys": {
            "description": ["Any examples of",
                            "Can you give me overview on",
                            "Give some examples of"]
        },

        "description": "what question (example)"
    },

    # what question (options)
    {
        "patterns": [
            "What <obj> {aux1} available",
            "What <obj> {aux2} <sbj> offers",
            "What <obj> {aux1} {offer} <sbj>",
            "What {aux1} <obj> {offer} <sbj>",
            "What kind of <obj> {aux2} <sbj> offer",

            "What are the options for <obj>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "offer": ["available in", "offered in", "offered by", 'available on']
        },

        "description": "what question (options)"
    },

    # what question (difference)
    {
        "patterns": [
            "What {aux1} {difference} <obj> {and} <obj>",
            "How {aux1} <obj> different",
            "How {aux2} <obj> differ from <obj>",
            "How {aux1} <obj> different from <obj>",
            "<st> how {aux2} <obj> fare compared to <obj>",
            "{difference} <obj> {and} <obj>",
            "<obj> {and} <obj> got what difference",
            "<obj> {and} <obj> got difference meh",
            "Compare <obj> {and} <obj>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "difference": ["the difference between",
                           "the main difference between",
                           "Difference between",
                           ],
            "and": ['and', '&', 'vs']
        },

        "description": "what question (difference)"
    },

    # comparison (similarity)
    {
        "patterns": [
            "{aux1} there any similarities between <obj> {and} <obj>",

            "How {aux2} <obj> related to <obj>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "and": ['and', '&', 'vs']
        },

        "description": "comparison (similarity)"
    },

    # what question (conditional)
    {
        "patterns": [
            "What {aux1} <obj> for <st>",
            "What <obj> {aux2} <sbj> {need} to <act> for <st>",

            "For <st> , what {aux1} <obj>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "need": ['need', 'have to do']
        },

        "description": "what question (conditional)"
    },

    # how to question
    {
        "patterns": [
            "How to <act>",
            "Where {aux1} <sbj> <act>",
            "Where {aux2} <sbj> <act>",
            "How {aux2} <sbj> <act>",
            "How {aux1} <sbj> <act>",
            "What {aux2} <sbj> {need} to <act>",
            "How <act>"
        ],
        "substitution_keys": {
            "aux1": ['can', 'could', 'should', 'shall'],
            "aux2": ['do', 'does'],
            "need": ['need', 'have to do']
        },

        "description": "how to question"
    },

    # how to question (role play)
    {
        "patterns": [
            "What {aux} <sbj> {need} as <obj> {while} <st>",
        ],
        "substitution_keys": {
            "aux": ['do', 'does'],
            "need": ['need to do', 'have to do'],
            "while": ['while', 'if']
        },

        "description": "how to question (role play)"
    },

    # how to question (conditional)
    {
        "patterns": [
            "<st> How {aux1} <sbj> <act>",
            "{if} <st> How {aux1} <sbj> <act>",
            "How to <act> if <st>",
            "<st> {description}",
            "{if} <st> {description}",
            "What to do if <st>",
            "What {aux1} <sbj> do if <st>",
            "<st> What {aux1} <sbj> do to <act>",
            "{if} <st> What {aux1} <sbj> do to <act>",
            "<st> Where {aux1} <sbj> <act>",
            "{if} <st> Where {aux1} <sbj> <act>",
            "<st> How {aux1} <sbj> <act>",
            "{if} <st> How {aux1} <sbj> <act>",
            "<st> How <sbj> {aux1} <act>",
            "{if} <st> How <sbj> {aux1} <act>",
            "<st> what {aux1} <sbj> do if <st>",
            "{if} <st> what {aux1} <sbj> do if <st>",
            "What {aux1} <sbj> do when <sbj> <act>",
            "{description} {if} <st>",
            "What {aux1} I do in the event <st>",
            "<st> I wish to <act> {description}",
            "{if} <st> I wish to <act> {description}",
            "<st> What {aux1} <sbj> do",
            "{if} <st> What {aux1} <sbj> do",
            "<st> Can you {advise} me on how to <act>",
            "{if} <st> Can you {advise} me on how to <act>",
            "<st> Can you {advise} me how to <act>",
            "{if} <st> Can you {advise} me how to <act>",
            "What else do <sbj> need to do to <act>",
            "<st> Could you help me <act>",
            "{if} <st> Could you help me <act>",
            "What {aux1} I do {if} <st>",
        ],
        "substitution_keys": {
            "aux1": ['can', 'could', 'should', 'shall', 'do', 'does'],
            "description": ["What is the procedure",
                            "What are your suggestions",
                            "What is your advice",
                            "Can you help me",
                            "Could you help me",
                            "What do I have to do",
                            "Can you tell me what to do",
                            "what can I do",
                            "Could you help"
                            ],
            "advise": ["advise", "guide", "please advise", "please guide"],
            "if": ['if', 'while', 'when']
        },

        "description": "how to question (conditional)"
    },

    # how to question (post-conditional)
    {
        "patterns": [
            "How {aux} <sbj> <act> after <st>",
            "What {aux} <sbj> {have} to do after <st>",
        ],
        "substitution_keys": {"aux": ['do', 'does'], "have": ['have', 'need']},

        "description": "how to question (post-conditional)"
    },

    # how stuff works question
    {
        "patterns": [
            "How {aux1} <obj> <act>",
            "How <obj> works",
            "How <obj> {aux1} <act>",
            "How {aux1} <obj> <act> for <obj>",
        ],
        "substitution_keys": {"aux1": ['is', 'are'], "aux2": ['do', 'does']},

        "description": "how stuff works question"
    },

    # how positive question
    {
        "patterns": [
            "How {good} {aux} <obj>",
            "{aux} <obj> {good}",
            "What {aux} so great about <obj>",
            "Can you tell me how {good} {aux} <obj>",
            "Explain how {good} {aux} <obj>"
        ],
        "substitution_keys": {
            "aux": ['is', 'are'],
            "good": ['good', 'well recognised', 'internationally recognised']
        },

        "description": "how positive question"
    },

    # how long question
    {
        "patterns": [
            "How long {aux3} <sbj> be <act>",
            "How long {aux2} <obj> <act>",
            "How long {aux3} <obj> <act>",
            "How long {aux1} <obj>",
            "What is the duration for <obj>",
            "How long <obj>"
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "aux3": ['can', 'should', 'will', 'would'],
        },

        "description": "how long question"
    },

    # how much question
    {
        "patterns": [
            "How {much} {aux1} <obj>",
            "How {much} {aux3} <sbj> <act>",
            "How {much} <obj> {aux2} <sbj> <act>",
            "How {much} <obj> {aux2} <sbj> {has}",
            "How {much} <obj>",
            "How {much} <obj> {aux1} there in <sbj>",
            "How {much} <obj> {aux1} there for <st>",

            "What is the amount of <obj>",
            "How {much} <obj>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does'],
            "aux3": ['can', 'should', 'will', 'would'],
            "much": ['much', 'many'],
            "has": ['has', 'have'],
        },

        "description": "how much question"
    },

    # why question I
    {
        "patterns": [
            "Why <act>",
            "<st> Why",
            "Why {aux} <sbj> <act>",
            "Why {aux} <st>",
            "<st> Why is {this} so",
            "Why is it that <st>",
            "<st> may I know why is {this} so",
            "Why <sbj> <act>",

            "<st> may I know why",
            "<st> what should I do"
        ],
        "substitution_keys": {
            "aux": ['am', 'is', 'are', 'do', 'does'],
            "this": ['this', 'that']
        },

        "description": "why question I"
    },

    # why question II
    {
        "patterns": [
            "Why {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['can', 'should', 'will', 'would']},

        "description": "why question II"
    },

    # why cannot question
    {
        "patterns": [
            "Why {aux} <sbj> {cannot} <act>",
            "Why <sbj> {cannot} <act> when <st>",
            "After <st> why {aux} <sbj> {cannot} <act>",
            "Why can not <sbj> <act>",

            "Why {cannot} <act>",
        ],
        "substitution_keys": {
            "aux": ['am', 'is', 'are'],
            "cannot": ['cannot', 'unable to']
        },

        "description": "why cannot question"
    },

    # where is question
    {
        "patterns": [
            "Where {aux1} <obj>",
            "Where {aux2} <sbj> find the procedure to <act>",
            "Where to <act>",
            "Where <obj>"
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['can', 'could']
        },

        "description": "where is question"
    },

    # who is question
    {
        "patterns": [
            "Who {aux2} <act>",
            "Who {aux1} <sbj>",
            "Who <act>",
            "Who {aux1} <st>",
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['can', 'could', 'should', 'shall', 'will', 'would'],
        },

        "description": "who is question"
    },

    # who does question
    {
        "patterns": [
            "Who {aux} <obj> <act>",
        ],
        "substitution_keys": {
            "aux": ['do', 'does']
        },

        "description": "who does question"
    },

    # who can I contact question
    {
        "patterns": [
            "Whom {aux2} <sbj> <act> regarding <obj>",
            "Who {aux2} <sbj> <act> regarding <obj>",
            "Who {aux1} <sbj> <act> if <st>",
            "Who {aux2} <sbj> <act> if <st>",
            "<st> who {aux2} <sbj> approach",
            "Who {aux2} I talk to about <obj>",
            "If <st> whom can <sbj> <act>",
            "Who {aux1} I <act> regarding <obj>",
            "Who {aux2} assist me if <st>",
            "Who {aux2} <sbj> contact regarding <obj>",
            "Who {aux2} <sbj> ask for assistance if <st>",
        ],
        "substitution_keys": {
            "aux1": ['do', 'does'],
            "aux2": ['can', 'could', 'should', 'shall']
        },

        "description": "who can I contact question"
    },

    # which question
    {
        "patterns": [
            "Which <obj> {aux} <act>",
            "Which <obj> <act>",
            "Which <obj>"
        ],
        "substitution_keys": {"aux": ['will', 'would', 'can', 'could']},

        "description": "which question"
    },

    # which to use question
    {
        "patterns": [
            "<st> {which} <obj> {aux1} <sbj> <act>",
            "{which} <obj> {aux1} <sbj> <act>",
            "<st> {which} {aux2} <obj>"
        ],
        "substitution_keys": {
            "aux1": ['do', 'does', 'can', 'could', 'should', 'shall'],
            "aux2": ['is', 'are'],
            "which": ['which', 'what']
        },

        "description": "which to use question"
    },

    # when question
    {
        "patterns": [
            "When {aux1} <obj>",
            "When {aux1} <obj> if <st>",
            "When {aux2} <sbj> <act>",
            "<st> When {aux2} <sbj> <act>",
            "When {aux3} be <obj>",
            "When <obj>"
        ],
        "substitution_keys": {
            "aux1": ['is', 'are'],
            "aux2": ['do', 'does', 'can', 'could', 'should', 'shall'],
            "aux3": ['will', 'would']
        },

        "description": "when question"
    },

    # when question (begin)
    {
        "patterns": [
            "What time {aux} <obj> {start}",
            "When {aux} <obj> {start}",
        ],
        "substitution_keys": {
            "aux": ['do', 'does', 'will', 'would'],
            "start": ['start', 'open'],
        },

        "description": "when question (begin)"
    },

    # when question (end)
    {
        "patterns": [
            "What time {aux} <obj> close",
            "Is there {deadline} for <obj>",
        ],
        "substitution_keys": {
            "aux": ['do', 'does', 'will', 'would'],
            "deadline": ['deadline', 'a deadline', 'closing date', 'a closing date'],
        },

        "description": "when question (end)"
    },

    # when question (obligatory)
    {
        "patterns": [
            "When {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['must', 'should', 'do', 'does']},

        "description": "when question (obligatory)"
    },

    # request question
    {
        "patterns": [
            "<st> {aux} you <act>",
            "{aux} you <act>",
        ],
        "substitution_keys": {
            "aux": ['can', 'could', 'will', 'would'],
        },

        "description": "request question"
    },

    # hypothetical question
    {
        "patterns": [
            "What if <sbj> <act>",
            "What if <st>",

            "If <st> then how",
            "If <sbj> <act> then how",
        ],
        "substitution_keys": {},

        "description": "hypothetical question"
    },

    # yes/no question (future tense)
    {
        "patterns": [
            "{aux} <sbj> <act> if <st>",
            "{aux} <sbj> <act>",
            "<st> {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "yes/no question (future tense)"
    },

    # can/cannot question (future tense)
    {
        "patterns": [
            "If <st> {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "can/cannot question (future tense)"
    },

    # what question (future tense)
    {
        "patterns": [
            "What <obj> {aux} <sbj> <act>",
            "I wish to know what <obj> {aux} be equivalent to <obj>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "what question (future tense)"
    },

    # how question (future tense)
    {
        "patterns": [
            "How {aux} <sbj> <act>",
            "<st> so how {aux} be <obj> <act>",
            "Under what circumstances {aux} <sbj> <act>",
            "How do <sbj> <act> if <st>",
            "<st> How {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "how question (future tense)"
    },

    # how many question (future tense)
    {
        "patterns": [
            "How {much} <obj> {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would'], "much": ['much', 'many']},

        "description": "how many question (future tense)"
    },

    # where question (future tense)
    {
        "patterns": [
            "Where {aux} <obj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "where question (future tense)"
    },

    # when question (future tense)
    {
        "patterns": [
            "If <st> when {aux} <sbj> <act>",
            "<st> when {aux} <sbj> <act>",
            "When {aux} <sbj> <act>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "when question (future tense)"
    },

    # prediction question (future tense)
    {
        "patterns": [
            "What {aux} happen to <obj>",
            "What {aux} happen to <obj> when <st>",
        ],
        "substitution_keys": {"aux": ['will', 'would']},

        "description": "prediction question (future tense)"
    },

    # when question (past tense)
    {
        "patterns": [
            "When {aux} <obj> <act>",

            "Since when {aux} <obj> <act>",
        ],
        "substitution_keys": {"aux": ['was', 'were']},

        "description": "when question (past tense)"
    }
]



import nlpaug
import nlpaug.augmenter.char as nac
import nlpaug.augmenter.word as naw
import re
import random
from tqdm import tqdm

class AUG():
    def __init__(self):
        #aug0 = naw.RandomWordAug()
        aug1 = naw.ContextualWordEmbsAug(model_path='bert-base-uncased', action="substitute")
        aug2 = naw.SynonymAug(aug_src='wordnet')
        #aug3 = naw.SplitAug()
        aug4 = naw.ContextualWordEmbsAug(model_path='bert-base-uncased', action="insert")

        self.augs = [aug1,aug2,aug4]
        

        
    
    def augment(self,sent : list ,n : int = 10) -> dict:
        ans = []
        sent = re.sub(r'[^a-zA-Z0-9_ ]', '', sent)
        for _ in range(n):
            aug = random.choice(self.augs)
            ans.append(aug.augment(sent))
                    
        return ans

    def exact_batch_generate(self,questions : list, n : int) -> dict:
        result = dict()

        for q in tqdm(questions):
            result[q] = self.augment(q, n)

        return result


import numpy as np


class BaseEncoder:
    def __init__(self, name, dimension, model_path):
        self.name = name
        self.dimension = dimension
        self.model_path = model_path

    def get_vector(self, sentence):
        raise NotImplementedError("'get_vector' is not implemented")

    def get_vectors(self, sentences):
        vectors = []
        for sentence in sentences:
            vectors.append(self.get_vector(sentence))

        return np.array(vectors)



import itertools
import string

from tqdm import tqdm

from  ...util import nlp
from .matcher import FuzzyMatcher
from .pattern import pattern_specs
from ..base import BaseGenerator

# 5W1H
WH = {'who', 'what', 'why', 'where', 'when', 'how'}
# Coordinating conjunction
FANBOYS = {'for', 'and', 'nor', 'but', 'or', 'yet', 'so'}


class FPMGenerator(BaseGenerator):
    """ Generate questions via fuzzy pattern matching on existing question patterns. """

    def __init__(self):
        super().__init__("Fuzzy Question Pattern Matching")

        self.patterns = self._generate_patterns(pattern_specs)
        self.matcher = FuzzyMatcher(self.patterns)

    @staticmethod
    def _generate_patterns(specs):
        patterns = []
        for group in specs:
            for _ in range(len(group['patterns'])):  # iterate through all original patterns
                pattern = group['patterns'].pop(0).strip()
                tokens_list = []  # list of tokens that require substitution
                for word in pattern.split():
                    if word.startswith('{') and word.endswith('}'):
                        tokens_list.append(word[1:-1])
                temp = [pattern]
                for token in tokens_list:
                    for _ in range(len(temp)):  # iterate through all sub-patterns
                        curr_pattern = temp.pop(0)
                        for substitute in group['substitution_keys'][token]:
                            temp.append(curr_pattern.replace('{' + token + '}', substitute))
                group['patterns'].extend(temp)

            patterns.append(group['patterns'])

        return patterns

    @staticmethod
    def _format_input(sentence):
        """ Add a space after specific punctuation if it is not followed by a space
        """
        punctuations = "!),.:;?]}"
        char_array = []
        for i, s in enumerate(sentence):
            char_array.append(s)
            if i < len(sentence) - 1 and s in punctuations and sentence[i + 1] not in punctuations + " ":
                char_array.append(" ")

        return "".join(char_array)

    @staticmethod
    def _format_output(sentence):
        words = sentence.strip("!,.:;?").split()
        words[0] = words[0][0].upper() + words[0][1:]  # capitalize first letter
        words[-1] = words[-1] + "?"

        return " ".join(words)

    @staticmethod
    def _has_multiple_question(question):
        """ Simple heuristic to determine whether a `question` contains multiple sub-questions.
        """
        if question.count('?') > 1:
            return True
        else:
            combined_wh = [["{} and {}".format(w1, w2), "{} & {}".format(w1, w2)] for w1 in WH for w2 in WH]
            return any([c in question.lower() for c in itertools.chain.from_iterable(combined_wh)])

    @staticmethod
    def _split_question(question):
        spacy_nlp = nlp.get_spacy_model()

        def _resolve_followup_question(prev_question, followup_question):
            """ Resolve sentences start with "If so", "If not", etc.
            """
            sub_question_lower = followup_question.lower()
            if any(sub_question_lower.startswith(pre) for pre in ['if so', 'if not']):
                last_aux_index = -1
                last_sub_index = -1
                index = 0
                with spacy_nlp.disable_pipes('ner'):
                    for token in spacy_nlp(prev_question):
                        if token.dep_ == 'aux':
                            last_aux_index = index
                        elif token.dep_ in ['nsubj', 'nsubjpass']:
                            last_sub_index = index
                        elif token.dep_ == 'ROOT':
                            break

                        if token.dep_ not in ['case', 'punct']:
                            index += 1

                tokens = prev_question.rstrip(".!?, ").split()
                sub = " ".join(tokens[last_aux_index + 1:last_sub_index + 1])
                aux = tokens[last_aux_index].lower()
                sub_sent = " ".join(tokens[last_sub_index + 1:]) + ","
                followup_question = " ".join(followup_question.split()[2:])
                if sub_question_lower.startswith('if so'):
                    followup_question = " ".join(["If", sub, aux, sub_sent, followup_question])
                elif sub_question_lower.startswith('if not'):
                    followup_question = " ".join(["If", sub, aux, "not", sub_sent, followup_question])

            return followup_question

        if not FPMGenerator._has_multiple_question(question):
            return [question]

        result = []
        with spacy_nlp.disable_pipes('ner'):
            doc = spacy_nlp(question)

            for sent in doc.sents:
                # remove leading coordinating conjunction
                for cc in FANBOYS:
                    if sent.text.lower().startswith(cc):
                        result.append(sent.text[len(cc) + 1:])
                        break
                else:
                    if len(result) >= 1 and any(sent.text.lower().startswith(pre) for pre in ['if so', 'if not']):
                        result.append(_resolve_followup_question(result[0], sent.text))
                    else:
                        result.append(sent.text)

        context = ""
        combined_wh = [["{} and {}".format(w1, w2), "{} & {}".format(w1, w2)] for w1 in WH for w2 in WH]
        for q in result.copy():
            if q.endswith('.'):
                context = q + ' '
                result.remove(q)
                continue

            for phrase in itertools.chain.from_iterable(combined_wh):
                if phrase in q.lower():
                    result.remove(q)
                    start_index = q.lower().index(phrase) + len(phrase)
                    sub_question = q[start_index:]
                    temp = phrase.split()
                    result.extend([temp[0].title() + sub_question, temp[-1].title() + sub_question])
                    break

        if context:
            results = [context + s for s in result]
        else:
            results = [nlp.resolve_coref(s, question) for s in result]

        return results

    @staticmethod
    def _preprocess(question):
        # 1. Convert contraction to regular form (e.g. "What's" to "What is")
        question = nlp.convert_contraction(FPMGenerator._format_input(question).strip())

        # 2. Separate multiple questions from sentence
        sentences = [question]
        if FPMGenerator._has_multiple_question(question):
            sentences = FPMGenerator._split_question(question)

        return sentences

    def generate(self, sentence):
        inputs = self._preprocess(sentence)

        result = []
        for sentence in inputs:
            # Extract question pattern
            matched_result = self.matcher.match(sentence)

            if not matched_result:
                continue
            else:
                tokens = matched_result.tokens
                if len(tokens['<st>']) > 1:
                    # concatenate multiple statements into single statement
                    tokens['<st>'] = [
                        ' and '.join(tokens['<st>']).translate(str.maketrans('', '', string.punctuation))]

                # Substitute tokens into other question patterns
                patterns_list = self.patterns[matched_result.group_id]
                for pattern in patterns_list:
                    if pattern.lower() == matched_result.pattern.lower():
                        continue
                    pattern_tokens = pattern.split()
                    if (len(tokens['<sbj>']) == pattern_tokens.count('<sbj>') and
                            len(tokens['<obj>']) >= pattern_tokens.count('<obj>') and
                            len(tokens['<act>']) == pattern_tokens.count('<act>') and
                            len(tokens['<st>']) >= pattern_tokens.count('<st>')):
                        for i in range(pattern_tokens.count('<obj>')):
                            pattern_tokens[pattern_tokens.index('<obj>')] = tokens['<obj>'][i].strip('?')
                        for i in range(pattern_tokens.count('<sbj>')):
                            pattern_tokens[pattern_tokens.index('<sbj>')] = tokens['<sbj>'][i].strip('?')
                        for i in range(pattern_tokens.count('<act>')):
                            pattern_tokens[pattern_tokens.index('<act>')] = tokens['<act>'][i].strip('?')
                        for i in range(pattern_tokens.count('<st>')):
                            pattern_tokens[pattern_tokens.index('<st>')] = tokens['<st>'][i].strip('?')

                        # Coreference resolution (i.e. Find out what 'it' in a sentence is referring to)
                        permuted = ' '.join(pattern_tokens)
                        if tokens['<st>'] and pattern.split().count('<st>') == 0:
                            result.append(self._format_output(
                                nlp.resolve_coref(permuted, sentence))
                            )
                        else:
                            result.append(self._format_output(permuted))

                result = list(set(result).difference({sentence}))

        return result

    def batch_generate(self, sentences):
        results = dict()
        for sentence in tqdm(sentences):
            results[sentence] = self.generate(sentence)

        return results



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "Bani",
    version = "0.2.1",
    author = "Rushi Babaria",
    url = "https://github.com/captanlevi/FAQ-QnA-matching.git",
    packages = setuptools.find_packages(),
    install_requires = ["sentence_transformers==0.3.8","pandas", "numpy","regex","nlpaug","spacy", "tqdm"],
    author_email="rjlinkin50@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6'

)




import json
import os
from pathlib import Path


def get_filename(path):
    return Path(path).name


def create_folder(path):
    if os.path.exists(path):
        print(f"{path} already exists.")
    else:
        os.mkdir(path)


def read_json(path):
    with open(path, 'r') as f:
        data = json.load(f)

    return data


def read_file(path, unique=False):
    print(f"Reading from {path}...")
    lines = []
    with open(path, 'r', encoding='utf-8') as f:
        unique_lines = set()
        for line in iter(f.readline, ''):
            line = line.strip()
            if unique and line not in unique_lines:
                unique_lines.add(line)
                lines.append(line)
            elif not unique:
                lines.append(line)

    print(f"Read {len(lines)} {'unique ' if unique else ''}lines from {path}")
    return lines


def write_file(lines, path):
    print(f"Writing {len(lines)} lines to {path}...")
    with open(path, "w", encoding='utf-8') as f:
        for line in lines:
            f.write(f"{line}\n")


def split_file(input_path, output_path1, output_path2, split_ratio):
    lines = read_file(input_path)
    num_lines = int(len(lines) * split_ratio)
    write_file(lines[num_lines:], output_path1)
    write_file(lines[:num_lines], output_path2)


def delete_file(*paths):
    for path in paths:
        os.remove(path)



from .eda import EDAGenerator
from .fpm.fpm import FPMGenerator
from .imt import IMTGenerator
from .symsub import SymSubGenerator
from .zeroshot import ZeroShotGenerator

__all__ = ["FPMGenerator", "SymSubGenerator", "IMTGenerator", "ZeroShotGenerator", "EDAGenerator"]



from logging import warning
from typing import Any, List, Union, Dict
import warnings



class QuestionGenerator:
    def __init__(self, name : str, producer, num : int):
        self.name = name

        if("exact_batch_generate" not in dir(producer) and "batch_generate" not in dir(producer)):
            raise AttributeError("Must have either batch_generate or exact batch generate in the producer object !!")
        self.producer = producer
        self.num = num

    def generate(self,questions: List[str]) -> Dict[str,List[str]]:
        """
        takes as input a list of questions

        the producer is an object that has a method , batch_generate

        batch generate takes in a list of questions as input and 
        outputs a dict .....
        the dict maps each orignal question to a list of generated questions
        """
        print("working with {} pipeline".format(self.name))
        if("exact_batch_generate" in dir(self.producer)):
            result_dict = self.producer.exact_batch_generate(questions,self.num)
        else:
            result_dict = self.producer.batch_generate(questions)
        if(result_dict is None):
            answer = dict()
            for q in questions:
                answer[q] = []
            return answer

        for orignal_question in result_dict:
            result_dict[orignal_question] = result_dict[orignal_question][:self.num]

        return result_dict



class IdentityProducer:
    """ 
    Use this generator when you do not need to use the generation pipeline !!!
    This will generate nothing and just return empty array !!!
    """
    def __init__(self):
        pass
    
    def batch_generate(self, questions: List[str]) -> Dict[str, List[str]]:
        resultDict = dict()
        for question in questions:
            resultDict[question] = []
        
        return resultDict




class GenerateManager:
    """
    Implement here functions that will help multiprocessing , moniter timing etc
    """
    def __init__(self , producers : List[Any], names : List[str] = None, nums : List[int] = None):
        """
        each producer is a user defined object that either implements 
        batch_generate or exact_batch_generate 
        """
        if(names is None):
            names = []
            for i in range(len(producers)):
                names.append("generator" + str(i))
        
        if(nums is None):
            nums = [10 for i in range(len(producers))]

        assert len(producers) == len(names) == len(nums)
        self.generators : List[QuestionGenerator] = [QuestionGenerator(name= names[i],producer= producers[i], num =  nums[i]) for i in range(len(producers))]


    def generate(self, questions : List[str]) -> List[List[str]]:
        """
        takes the questions given , and uses the generators to generate.
        Returns generated questions in order !!!
        """

        mergedDict = self._sequential(questions)

        result = []

        for question in questions:
            if(question in mergedDict):
                result.append(mergedDict[question])
            else:
                warnings.warn("Zero questions generated for ---> {}".format(question))
                result.append([])

        return result

    
    def _mergeDicts(self, dicts : List[Dict[str , List[str]]]) -> Dict[str, List[str]]:
        mergedDict = dict()
        for dct in dicts:
            for que,gen_ques in dct.items():
                if(que not in mergedDict):
                    mergedDict[que] = []
                mergedDict[que].extend(gen_ques)
            
    
        return mergedDict

    

    def _sequential(self, questions):
        """
        Applies generators in a series , ie one after another no 
        multiprocessing or multithreading 
        """

        resultDicts = []

        for generator in self.generators:
            try:
                resultDicts.append(generator.generate(questions))
            except Exception as exception:
                warnings.warn("{} pipeline failed !!!! {}".format(generator.name,exception))
        
        return self._mergeDicts(resultDicts)









